#ifndef SCHOENHAGE_H_INCLUDED
#define SCHOENHAGE_H_INCLUDED

#include <gmp.h>

/* In limbs */
#define SCHOENHAGE_LIMIT 2

struct schoenhage_array
{
  mpz_t m[2][2];
};

/* Helper functions */

int
normalize(mpz_t r, mpz_t a, mpz_t b);

void
fix(struct schoenhage_array *R, mpz_t x, mpz_t y);

void
schoenhage_array_init(struct schoenhage_array *R);

void
schoenhage_array_clear(struct schoenhage_array *R);

void
schoenhage_array_identity(struct schoenhage_array *R);

void
schoenhage_array_copy(struct schoenhage_array *R,
		      const struct schoenhage_array *S);

int
schoenhage_array_equal(const struct schoenhage_array *R,
		       const struct schoenhage_array *S);
int
schoenhage_array_row_equal(const struct schoenhage_array *R,
			   const struct schoenhage_array *S);

void
schoenhage_array_step_ui(struct schoenhage_array *R,
			 mp_limb_t q);

void
schoenhage_array_step_left(struct schoenhage_array *R,
			   const mpz_t q);

void
schoenhage_array_step(struct schoenhage_array *R,
		      const mpz_t q);

void
schoenhage_array_vector(mpz_t y1, mpz_t y2,
			const struct schoenhage_array *R,
			const mpz_t x1, const mpz_t x2);

void
schoenhage_array_mult(struct schoenhage_array *R,
		      const struct schoenhage_array *A,
		      const struct schoenhage_array *B);

void
egcd(struct schoenhage_array *R, unsigned size, mp_limb_t a, mp_limb_t b);

void
check_hgcd(const struct schoenhage_array *R,
	   const mpz_t a, const mpz_t b);

int
hgcd(struct schoenhage_array *R, const mpz_t a, const mpz_t b);

int
hgcd_slow(struct schoenhage_array *R, const mpz_t a, const mpz_t b);

void
fgcd(mpz_t r, mpz_t a, mpz_t b);

void
schoenhage(mpz_t r, const mpz_t x, const mpz_t y);

#define trace_array(msg, R) \
trace("%s: [ %Zd, %Zd\n   %Zd, %Zd\n", (msg), \
  (R)->m[0][0], (R)->m[0][1], (R)->m[1][0], (R)->m[1][1])



#endif /* SCHOENHAGE_H_INCLUDED */
